/*
The copyright in this software is being made available under the BSD License, included below. This software may be subject to other third party and contributor rights, including patent rights, and no such rights are granted under this license.
 
Copyright (c) 2013, ITU and ISO/IEC
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name of the ITU and ISO/IEC nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include "yuv_proc.h"
#include <string.h>
#pragma warning( disable: 4996 )

inline byte ClipToByte( int iX ) {return (byte)( iX > 0 ? ( iX < 255? iX : 255 ) : 0 ); }

void YUV444To420Then444( char *InputName, char *OutName, int srcH, int srcW ) {
    FILE *pSrcFile = fopen( InputName, "rb" );
    FILE *pDstFile = fopen( OutName, "wb" );
    byte *ucTempFrame = new byte[srcH * srcW];
    byte *ucpTemp;
    //while ( !feof( pSrcFile ) ) {
    for ( int i = 0; i < 57 ; ++ i ) {
        fread( ucTempFrame, sizeof(byte), srcH * srcW, pSrcFile );
        fwrite( ucTempFrame, sizeof(byte), srcH * srcW, pDstFile );
        fread( ucTempFrame, sizeof(byte), srcH * srcW, pSrcFile );
        ucpTemp = ucTempFrame + srcW;
        for ( int y = 1; y < srcH - 1; y += 2 ) {
            for ( int x = 0; x < srcW; x += 2 ) {
                *(ucpTemp + x) = ( *(ucpTemp + srcW + x) + *(ucpTemp - srcW + x) + 1 ) >> 1;
            }
            ucpTemp += (srcW * 2);
        }
        ucpTemp = ucTempFrame;
        for ( int y = 0; y < srcH; y += 2 ) {
            for ( int x = 1; x < srcW - 1; x += 2 ) {
                *(ucpTemp + x) = ( *(ucpTemp + x - 1) + *(ucpTemp + x + 1) + 1 ) >> 1;
            }
            *(ucpTemp + srcW - 1) = *(ucpTemp + srcW - 2);
            ucpTemp += (srcW * 2);
        }
        ucpTemp = ucTempFrame + srcW;
        for ( int y = 1; y < srcH - 1; y += 2 ) {
            for ( int x = 1; x < srcW - 1; x += 2 ) {
                *(ucpTemp + x) = ( *(ucpTemp + srcW + x - 1) + *(ucpTemp + srcW + x + 1) + *(ucpTemp - srcW + x - 1) + *(ucpTemp - srcW + x + 1) + 2 ) >> 2;
            }
            *(ucpTemp + srcW - 1) = *(ucpTemp + srcW - 2);
            ucpTemp += (srcW * 2);
        }
        memcpy( ucpTemp, ucTempFrame, srcW * sizeof(byte));
        fwrite( ucTempFrame, sizeof(byte), srcH * srcW, pDstFile );
        fread( ucTempFrame, sizeof(byte), srcH * srcW, pSrcFile );
        ucpTemp = ucTempFrame + srcW;
        for ( int y = 1; y < srcH - 1; y += 2 ) {
            for ( int x = 0; x < srcW; x += 2 ) {
                *(ucpTemp + x) = ( *(ucpTemp + srcW + x) + *(ucpTemp - srcW + x) + 1 ) >> 1;
            }
            ucpTemp += (srcW * 2);
        }
        ucpTemp = ucTempFrame;
        for ( int y = 0; y < srcH; y += 2 ) {
            for ( int x = 1; x < srcW - 1; x += 2 ) {
                *(ucpTemp + x) = ( *(ucpTemp + x - 1) + *(ucpTemp + x + 1) + 1 ) >> 1;
            }
            *(ucpTemp + srcW - 1) = *(ucpTemp + srcW - 2);
            ucpTemp += (srcW * 2);
        }
        ucpTemp = ucTempFrame + srcW;
        for ( int y = 1; y < srcH - 1; y += 2 ) {
            for ( int x = 1; x < srcW - 1; x += 2 ) {
                *(ucpTemp + x) = ( *(ucpTemp + srcW + x - 1) + *(ucpTemp + srcW + x + 1) + *(ucpTemp - srcW + x - 1) + *(ucpTemp - srcW + x + 1) + 2 ) >> 2;
            }
            *(ucpTemp + srcW - 1) = *(ucpTemp + srcW - 2);
            ucpTemp += (srcW * 2);
        }
        memcpy( ucpTemp, ucTempFrame, srcW * sizeof(byte));
        fwrite( ucTempFrame, sizeof(byte), srcH * srcW, pDstFile );
    }
    delete [] ucTempFrame;
    fclose( pSrcFile );
    fclose( pDstFile );
}

void YUVCutShort( char *InputName, char *OutName, int srcH, int srcW, int iOffset, int iNumFrames, YUVFormat eFormat ) {
    FILE *pSrcFile = fopen( InputName, "rb" );
    FILE *pDstFile = fopen( OutName, "wb" );
    byte *ucTempFrame;
    if ( eFormat == RGB_32 ) {
        ucTempFrame = new byte[srcH * srcW * 4];
        for ( int i = 0; i < iOffset; ++ i ) {
            fread( ucTempFrame, sizeof(byte), ( srcH * srcW ) * 4, pSrcFile );
        }
        for ( int i = 0; i < iNumFrames; ++ i ) {
            fread( ucTempFrame, sizeof(byte), ( srcH * srcW ) * 4, pSrcFile );
            fwrite( ucTempFrame, sizeof(byte), ( srcH * srcW ) * 4, pDstFile );
        }
    } else {
        ucTempFrame = new byte[(srcH * srcW * 3) >> 1];
        for ( int i = 0; i < iNumFrames; ++ i ) {
            fread( ucTempFrame, sizeof(byte), ( srcH * srcW ) * 3 >> 1, pSrcFile );
            fwrite( ucTempFrame, sizeof(byte), ( srcH * srcW * 3 ) >> 1, pDstFile );
        }
    }
    delete [] ucTempFrame;
    fclose( pSrcFile );
    fclose( pDstFile );
}

void YUVConversion( char *InputName, char *OutName, int srcH, int srcW, YUVFormat InFormat, YUVFormat OutFormat ) {
    if ( InFormat == YUV_444 ) {
        if ( OutFormat == RGB_32 ) {
            YUV444ToRGB32( InputName, OutName, srcH, srcW );
        }
    }
}

void YUV444ToRGB32( char *InputName, char *OutName, int srcH, int srcW ) {
    FILE *pSrcFile = fopen( InputName, "rb" );
    FILE *pDstFile = fopen( OutName, "wb" );
    int y, x;
    byte *YLine, *ULine, *VLine;
    byte *Y = new byte[srcH * srcW];
    byte *U = new byte[srcH * srcW];
    byte *V = new byte[srcH * srcW];
    byte *BGRLine = new byte[srcW * 4];
    int iCnt = 0;
    printf( "Finished:      ", iCnt );
    while ( !feof( pSrcFile ) )
    {
        fread( Y, sizeof(byte), srcH * srcW, pSrcFile );
        if ( feof( pSrcFile ) )
        {
            break;
        }
        fread( U, sizeof(byte), srcH * srcW, pSrcFile );
        fread( V, sizeof(byte), srcH * srcW, pSrcFile );
        YLine = Y; ULine = U; VLine = V;
        for ( y = 0; y < srcH; ++ y ) 
        {
            for ( x = 0; x < srcW; ++ x )
            {
                InverseColorConversion( *( YLine + x ), *( ULine + x ), *( VLine + x ), (BGRLine + 4 * x + 2 ), (BGRLine + 4 * x + 1 ), (BGRLine + 4 * x ) );
            }
            YLine += srcW;
            ULine += srcW;
            VLine += srcW;
            fwrite( BGRLine, sizeof(byte), srcW * 4, pDstFile );
        }
        iCnt ++;
        printf( "\b\b\b\b\b%04d.", iCnt );
    }
    printf( "\n" );
    fclose( pSrcFile );
    fclose( pDstFile );
    delete [] Y;
    delete [] U;
    delete [] V;
    delete [] BGRLine;
}

void YUVCutSequence( char *InputName, char *OutName, int srcH, int srcW, YUVFormat InFormat ) {
    FILE *pSrcFile = fopen( InputName, "rb" );
    FILE *pDstFile = fopen( OutName, "wb" );
    byte *Y = new byte[srcH * srcW * 3 / 2];
    for ( int i = 0; i < 50; ++ i ) {
        fread( Y, sizeof(byte), srcH * srcW * 3 / 2, pSrcFile );
        fwrite( Y, sizeof(byte), srcH * srcW * 3 / 2, pDstFile );
    }
    delete [] Y;
    fclose( pSrcFile );
    fclose( pDstFile );
}

void YUVChangeFrameWithColor( char *InputName, char *OutName, int srcH, int srcW, YUVFormat InFormat ) {
    FILE *pSrcFile = fopen( InputName, "rb" );
    FILE *pDstFile = fopen( OutName, "wb" );
    byte *Y = new byte[srcH * srcW * 3 / 2];
    
    int p1 = 18;
    int p2 = 20;
    int p3 = 21;
    int p4 = 22;
    int p5 = 50;

    for ( int i = 0; i < p1; i ++ ) {
        fread( Y, sizeof(byte), srcH * srcW * 3 / 2, pSrcFile );
        fwrite( Y, sizeof(byte), srcH * srcW * 3 / 2, pDstFile );
    }
    for ( int i = 0; i < srcH * srcW *3 / 2; ++ i ) {
        Y[i] = 128;
    }
    fseek( pSrcFile, srcH *srcW * 3 / 2, SEEK_CUR );
    fwrite( Y, sizeof(byte), srcH * srcW * 3 / 2, pDstFile );
    for ( int i = p1 + 1; i < p5; i ++ ) {
        fread( Y, sizeof(byte), srcH * srcW * 3 / 2, pSrcFile );
        fwrite( Y, sizeof(byte), srcH * srcW * 3 / 2, pDstFile );
    }

    
    fclose( pSrcFile );
    fclose( pDstFile );
    delete [] Y;
}

void FramePacking444( char *InputName, char *OutName, int srcH, int srcW, double dAlpha, double dBeta ) 
{
    FILE *pSrcFile = fopen( InputName, "rb" );
    FILE *pDstFile = fopen( OutName, "wb" );
    byte *pSrcYUV444 = new byte[srcH * srcW * 3];
    byte *pDstY = new byte[srcH * srcW * 2];
    byte *pDstU = new byte[srcH * srcW / 2];
    byte *pDstV = new byte[srcH * srcW / 2];
    int iDstStride = srcW;
    int iDstStrideUV = srcW / 2;
    

    int x, y;
    byte Y, U, V;
    int iCnt = 0;
    printf( "Finished:      " );
    while ( !feof( pSrcFile ) )
    {
        byte *pSrcRGB       = pSrcYUV444;
        byte *pDestY420		= pDstY;
        byte *pDestU420		= pDstU;
        byte *pDestV420		= pDstV;
        byte *pDestHalfU    = pDstY+iDstStride*srcH;
        byte *pDestHalfV	= pDstY+iDstStride*srcH*3/2;
        byte *pDestEighthU1	= pDstU+iDstStrideUV*srcH/2;
        byte *pDestEighthV1	= pDstU+iDstStrideUV*srcH*3/4;
        byte *pDestEighthU2	= pDstV+iDstStrideUV*srcH/2;
        byte *pDestEighthV2	= pDstV+iDstStrideUV*srcH*3/4;
        fread( pSrcYUV444, sizeof(byte), srcH * srcW * 3, pSrcFile );
        if ( feof( pSrcFile ) ) 
        {
            break;
        }
        Y = *pSrcYUV444;
        U = *(pSrcYUV444 + srcH * srcW);
        V = *(pSrcYUV444 + srcH * srcW * 2);
        for ( y = 0; y < srcH; ++ y ) 
        {
            for ( x = 0; x < srcW; ++ x ) 
            {
                *(pDestY420 + x) = Y; 
                if ( y & 1 ) {
                    *(pDestHalfU+x) = U;
                    *(pDestHalfV+x) = V;
                } else if ( 0==(x & 1) ) {
                    *(pDestU420+x/2) = U;
                    *(pDestV420+x/2) = V;
                } else if ( 1== (x & 1 ) && 0==(y & 2) ) {
                    *(pDestEighthU1+x/2) = U;
                    *(pDestEighthV1+x/2) = V;
                } else { // if ( 1== (x & 1 ) && (y & 2) ) 
                    *(pDestEighthU2+x/2) = U;
                    *(pDestEighthV2+x/2) = V;
                }
            }
            Y ++;
            U ++;
            V ++;
            pSrcRGB += iDstStride * 4;
            pDestY420 += iDstStride;
            if ( y & 1 ) {
                pDestHalfU  += iDstStride;
                pDestHalfV  += iDstStride;
            } else  {
                pDestU420 += iDstStrideUV;
                pDestV420 += iDstStrideUV;
                if ( 0== ( y & 2 ) ) {
                    pDestEighthU1 += iDstStrideUV;
                    pDestEighthV1 += iDstStrideUV;
                }
                else {
                    pDestEighthU2 += iDstStrideUV;
                    pDestEighthV2 += iDstStrideUV;
                }
            }
        }
        pDestY420		= pDstY;
        pDestU420		= pDstU;
        pDestV420		= pDstV;
        pDestHalfU    = pDstY+iDstStride*srcH;
        pDestHalfV	= pDstY+iDstStride*srcH*3/2;
        pDestEighthU1	= pDstU+iDstStrideUV*srcH/2;
        pDestEighthV1	= pDstU+iDstStrideUV*srcH*3/4;
        pDestEighthU2	= pDstV+iDstStrideUV*srcH/2;
        pDestEighthV2	= pDstV+iDstStrideUV*srcH*3/4;
        double a, b, c, d, tmp;
        for ( y = 0; y < srcH / 2 ; ++ y ) 
        {
            for ( x = 0; x < srcW / 2; ++ x ) 
            {
                if ( y & 1 ) {
                    b = (double)(*(pDestEighthU2 + iDstStrideUV * (y / 2 ) + x ));
                } else {
                    b = (double)(*(pDestEighthU1 + iDstStrideUV * (y / 2 ) + x ));
                }
                a = (double)(*(pDestU420 + y * iDstStrideUV + x));
                d = (double)(*(pDestHalfU + y * iDstStride + x * 2 + 1 ));
                c = (double)(*(pDestHalfU + y * iDstStride + x * 2 ));
                tmp = (( a + dAlpha * (b + c) + dBeta * d ) / ( 1 + dAlpha * 2 + dBeta ) + 0.5 );
                tmp = tmp > 255.0 ? 255 : tmp;
                tmp = tmp < 0.0 ? 0 : tmp;
                *(pDestU420 + y * iDstStrideUV + x) = (byte)tmp;
                if ( y & 1 ) {
                    b = (double)(*(pDestEighthV2 + iDstStrideUV * (y / 2 ) + x ));
                } else {
                    b = (double)(*(pDestEighthV1 + iDstStrideUV * (y / 2 ) + x ));
                }
                a = (double)(*(pDestV420 + y * iDstStrideUV + x));
                d = (double)(*(pDestHalfV + y * iDstStride + x * 2 + 1 ));
                c = (double)(*(pDestHalfV + y * iDstStride + x * 2 ));
                tmp = (( a + dAlpha * (b + c) + dBeta * d ) / ( 1 + dAlpha * 2 + dBeta ) + 0.5 );
                tmp = tmp > 255.0 ? 255 : tmp;
                tmp = tmp < 0.0 ? 0 : tmp;
                *(pDestV420 + y * iDstStrideUV + x) = (byte)tmp;
            }
        }
        fwrite( pDstY, sizeof(byte), srcH * srcW * 2, pDstFile );
        fwrite( pDstU, sizeof(byte), srcH * srcW / 2, pDstFile );
        fwrite( pDstV, sizeof(byte), srcH * srcW / 2, pDstFile );
        iCnt ++;
        printf( "\b\b\b\b\b%04d.", iCnt );
    }
    printf( "\n" );
    delete[] pSrcYUV444;
    delete[] pDstY;
    delete[] pDstU;
    delete[] pDstV;
    fclose( pSrcFile );
    fclose( pDstFile );
}

byte Round( double dIn )
{
    int iOut = (int)( dIn + 0.5 );
    return (byte)((iOut < 0) ? 0 : ((iOut > 255) ? 255 : iOut));
}

void ColorConversion( byte R, byte G, byte B, byte* Y, byte* U, byte* V )
{
#ifdef BT601
    double Kr = 0.299;
    double Kb = 0.114;
#else
    double Kr = 0.2126;
    double Kb = 0.0722;
#endif
    double Kg = 1.0 - Kr - Kb;
    double Dcb = ( 2.0 *  ( 1.0 - Kb ) );
    double Dcr = ( 2.0 *  ( 1.0 - Kr ) );


    double Ey  = ( Kr * R + Kg * G + Kb * B ) / 255.0;
    double Epb = ( B / 255.0 - Ey ) / Dcb;
    double Epr = ( R / 255.0 - Ey ) / Dcr;

    *Y = Round( 219.0 * Ey + 16.0 );
    *U = Round( 224.0 * Epb + 128.0 );
    *V = Round( 224.0 * Epr + 128.0 );
}

void InverseColorConversion( byte Y, byte U, byte V, byte* R, byte* G, byte* B )
{
#ifdef BT601
    double Kr = 0.299;
    double Kb = 0.114;
#else
    double Kr = 0.2126;
    double Kb = 0.0722;
#endif
    double Kg = 1.0 - Kr - Kb;
    double Dcb = ( 2.0 *  ( 1.0 - Kb ) );
    double Dcr = ( 2.0 *  ( 1.0 - Kr ) );

    double Ey  = ( Y - 16.0 ) / 219.0;
    double Epb = ( U - 128.0 ) / 224.0;
    double Epr = ( V - 128.0 ) / 224.0;

    double Fr = 255.0 * ( Dcr * Epr + Ey );
    double Fb = 255.0 * ( Dcb * Epb + Ey );
    double Fg = ( ( 255.0 * Ey ) - Fr * Kr - Kb * Fb ) / Kg;

    *R = Round( Fr );
    *G = Round( Fg );
    *B = Round( Fb );
}

void FramePacking( char *InputName, char *OutName, int srcH, int srcW, double dAlpha, double dBeta, FileType eFileType )
{
    FILE *pSrcFile = fopen( InputName, "rb" );
    FILE *pDstFile = fopen( OutName, "wb" );
    int iFactorFileType;
    if ( RGB32 == eFileType )
    {
        iFactorFileType = 4;
    }
    else
    {
        iFactorFileType = 3;
    }
    byte *pSrc = new byte[srcH * srcW * iFactorFileType];
    byte *pDstY = new byte[srcH * srcW * 2];
    byte *pDstU = new byte[srcH * srcW / 2];
    byte *pDstV = new byte[srcH * srcW / 2];
    int iDstStride = srcW;
    int iDstStrideUV = srcW / 2;
    

    int x, y, R, G, B;
    byte Y, U, V;
    int iCnt = 0;
    printf( "Finished:      " );
    while ( !feof( pSrcFile ) )
    {
        byte *pSrcRGB       = pSrc;
        byte *pDestY420		= pDstY;
        byte *pDestU420		= pDstU;
        byte *pDestV420		= pDstV;
        byte *pDestHalfU    = pDstY+iDstStride*srcH;
        byte *pDestHalfV	= pDstY+iDstStride*srcH*3/2;
        byte *pDestEighthU1	= pDstU+iDstStrideUV*srcH/2;
        byte *pDestEighthV1	= pDstU+iDstStrideUV*srcH*3/4;
        byte *pDestEighthU2	= pDstV+iDstStrideUV*srcH/2;
        byte *pDestEighthV2	= pDstV+iDstStrideUV*srcH*3/4;
        fread( pSrc, sizeof(byte), srcH * srcW * iFactorFileType, pSrcFile );
        if ( feof( pSrcFile ) ) 
        {
            break;
        }
        byte* pY = pSrc;
        byte* pU = pSrc + srcH * srcW;
        byte* pV = pSrc + srcH * srcW * 2;
        for ( y = 0; y < srcH; ++ y ) 
        {
            for ( x = 0; x < srcW; ++ x ) 
            {
                if ( RGB32 == eFileType )
                {
                    B =  pSrcRGB[4*x + 0];
                    G =  pSrcRGB[4*x + 1];
                    R =  pSrcRGB[4*x + 2];
                    ColorConversion( R, G, B, &Y, &U, &V );
                }
                else
                {
                    Y = *pY;
                    U = *pU;
                    V = *pV;
                    pY ++;
                    pU ++;
                    pV ++;
                }
                *(pDestY420 + x) = Y; 
                if ( y & 1 ) {
                    *(pDestHalfU+x) = U;
                    *(pDestHalfV+x) = V;
                } else if ( 0==(x & 1) ) {
                    *(pDestU420+x/2) = U;
                    *(pDestV420+x/2) = V;
                } else if ( 1== (x & 1 ) && 0==(y & 2) ) {
                    *(pDestEighthU1+x/2) = U;
                    *(pDestEighthV1+x/2) = V;
                } else { // if ( 1== (x & 1 ) && (y & 2) ) 
                    *(pDestEighthU2+x/2) = U;
                    *(pDestEighthV2+x/2) = V;
                }
            }
            pSrcRGB += iDstStride * 4;
            pDestY420 += iDstStride;
            if ( y & 1 ) {
                pDestHalfU  += iDstStride;
                pDestHalfV  += iDstStride;
            } else  {
                pDestU420 += iDstStrideUV;
                pDestV420 += iDstStrideUV;
                if ( 0== ( y & 2 ) ) {
                    pDestEighthU1 += iDstStrideUV;
                    pDestEighthV1 += iDstStrideUV;
                }
                else {
                    pDestEighthU2 += iDstStrideUV;
                    pDestEighthV2 += iDstStrideUV;
                }
            }
        }
        pDestY420		= pDstY;
        pDestU420		= pDstU;
        pDestV420		= pDstV;
        pDestHalfU    = pDstY+iDstStride*srcH;
        pDestHalfV	= pDstY+iDstStride*srcH*3/2;
        pDestEighthU1	= pDstU+iDstStrideUV*srcH/2;
        pDestEighthV1	= pDstU+iDstStrideUV*srcH*3/4;
        pDestEighthU2	= pDstV+iDstStrideUV*srcH/2;
        pDestEighthV2	= pDstV+iDstStrideUV*srcH*3/4;
        double a, b, c, d, tmp;
        for ( y = 0; y < srcH / 2 ; ++ y ) {
            for ( x = 0; x < srcW / 2; ++ x ) {
                if ( y & 1 ) {
                    b = (double)(*(pDestEighthU2 + iDstStrideUV * (y / 2 ) + x ));
                } else {
                    b = (double)(*(pDestEighthU1 + iDstStrideUV * (y / 2 ) + x ));
                }
                a = (double)(*(pDestU420 + y * iDstStrideUV + x));
                d = (double)(*(pDestHalfU + y * iDstStride + x * 2 + 1 ));
                c = (double)(*(pDestHalfU + y * iDstStride + x * 2 ));
                tmp = (( a + dAlpha * (b + c) + dBeta * d ) / ( 1 + dAlpha * 2 + dBeta ) + 0.5 );
                tmp = tmp > 255.0 ? 255 : tmp;
                tmp = tmp < 0.0 ? 0 : tmp;
                *(pDestU420 + y * iDstStrideUV + x) = (byte)tmp;
                if ( y & 1 ) {
                    b = (double)(*(pDestEighthV2 + iDstStrideUV * (y / 2 ) + x ));
                } else {
                    b = (double)(*(pDestEighthV1 + iDstStrideUV * (y / 2 ) + x ));
                }
                a = (double)(*(pDestV420 + y * iDstStrideUV + x));
                d = (double)(*(pDestHalfV + y * iDstStride + x * 2 + 1 ));
                c = (double)(*(pDestHalfV + y * iDstStride + x * 2 ));
                tmp = (( a + dAlpha * (b + c) + dBeta * d ) / ( 1 + dAlpha * 2 + dBeta ) + 0.5 );
                tmp = tmp > 255.0 ? 255 : tmp;
                tmp = tmp < 0.0 ? 0 : tmp;
                *(pDestV420 + y * iDstStrideUV + x) = (byte)tmp;
            }
        }
        fwrite( pDstY, sizeof(byte), srcH * srcW * 2, pDstFile );
        fwrite( pDstU, sizeof(byte), srcH * srcW / 2, pDstFile );
        fwrite( pDstV, sizeof(byte), srcH * srcW / 2, pDstFile );
        iCnt ++;
        printf( "\b\b\b\b\b%04d.", iCnt );
    }
    printf( "\n" );
    delete[] pSrc;
    delete[] pDstY;
    delete[] pDstU;
    delete[] pDstV;
    fclose( pSrcFile );
    fclose( pDstFile );
}


void FramePackingWithBandSeparation( char *InputName, char *OutName, int srcH, int srcW, double dAlpha, double dBeta, FileType eFileType ) 
{
    FILE *pSrcFile = fopen( InputName, "rb" );
    FILE *pDstFile = fopen( OutName, "wb" );
    int iFactorFileType;
    if ( RGB32 == eFileType )
    {
        iFactorFileType = 4;
    }
    else
    {
        iFactorFileType = 3;
    }
    byte *pSrc = new byte[srcH * srcW * iFactorFileType];
    byte *pDstY = new byte[srcH * srcW * 2];
    byte *pDstU = new byte[srcH * srcW / 2];
    byte *pDstV = new byte[srcH * srcW / 2];
    int iDstStride = srcW;
    int iDstStrideUV = srcW / 2;
    int x, y, R, G, B;
    byte Y, U, V;
    int **ppCU, **ppCV, **ppDU, **ppDV;
    ppCU = new int*[srcH/ 2];
    ppDU = new int*[srcH / 2];
    ppCV = new int*[srcH / 2];
    ppDV = new int*[srcH / 2];
    for ( y = 0; y < srcH / 2; ++ y )
    {
        ppCU[y] = new int[srcW];
        ppCV[y] = new int[srcW];
        ppDU[y] = new int[srcW];
        ppDV[y] = new int[srcW];
    }
    int iCnt = 0;
    printf( "Finished:      " );
    while ( !feof( pSrcFile ) )
    {
        byte *pSrcRGB       = pSrc;
        byte *pDestY420		= pDstY;
        byte *pDestU420		= pDstU;
        byte *pDestV420		= pDstV;
        byte *pDestHalfU    = pDstY+iDstStride*srcH;
        byte *pDestHalfV	= pDstY+iDstStride*srcH*3/2;
        byte *pDestEighthU1	= pDstU+iDstStrideUV*srcH/2;
        byte *pDestEighthV1	= pDstU+iDstStrideUV*srcH*3/4;
        byte *pDestEighthU2	= pDstV+iDstStrideUV*srcH/2;
        byte *pDestEighthV2	= pDstV+iDstStrideUV*srcH*3/4;
        fread( pSrc, sizeof(byte), srcH * srcW * iFactorFileType, pSrcFile );
        if ( feof( pSrcFile ) )
        {
            break;
        }
        byte* pY = pSrc;
        byte* pU = pSrc + srcH * srcW;
        byte* pV = pSrc + srcH * srcW * 2;
        for ( y = 0; y < srcH; ++ y ) 
        {
            for ( x = 0; x < srcW; ++ x ) 
            {
                if ( RGB32 == eFileType )
                {
                    B =  pSrcRGB[4*x + 0];
                    G =  pSrcRGB[4*x + 1];
                    R =  pSrcRGB[4*x + 2];
                    ColorConversion( R, G, B, &Y, &U, &V );
                }
                else
                {
                    Y = *pY;
                    U = *pU;
                    V = *pV;
                    pY ++;
                    pU ++;
                    pV ++;
                }
                *(pDestY420 + x) = Y; 
                if ( y & 1 ) {
                    *(pDestHalfU+x) = U;
                    *(pDestHalfV+x) = V;
                } else if ( 0==(x & 1) ) {
                    *(pDestU420+x/2) = U;
                    *(pDestV420+x/2) = V;
                } else if ( 1== (x & 1 ) && 0==(y & 2) ) {
                    *(pDestEighthU1+x/2) = U;
                    *(pDestEighthV1+x/2) = V;
                } else { // if ( 1== (x & 1 ) && (y & 2) ) 
                    *(pDestEighthU2+x/2) = U;
                    *(pDestEighthV2+x/2) = V;
                }
            }
            pSrcRGB += iDstStride * 4;
            pDestY420 += iDstStride;
            if ( y & 1 ) {
                pDestHalfU  += iDstStride;
                pDestHalfV  += iDstStride;
            } else  {
                pDestU420 += iDstStrideUV;
                pDestV420 += iDstStrideUV;
                if ( 0== ( y & 2 ) ) {
                    pDestEighthU1 += iDstStrideUV;
                    pDestEighthV1 += iDstStrideUV;
                }
                else {
                    pDestEighthU2 += iDstStrideUV;
                    pDestEighthV2 += iDstStrideUV;
                }
            }
        }
        // prefiltering
        pDestY420		= pDstY;
        pDestU420		= pDstU;
        pDestV420		= pDstV;
        pDestHalfU		= pDstY+iDstStride*srcH;
        pDestHalfV		= pDstY+iDstStride*srcH*3/2;
        pDestEighthU1	= pDstU+iDstStrideUV*srcH/2;
        pDestEighthV1	= pDstU+iDstStrideUV*srcH*3/4;
        pDestEighthU2	= pDstV+iDstStrideUV*srcH/2;
        pDestEighthV2	= pDstV+iDstStrideUV*srcH*3/4;
        double a, b, c, d, tmp;
        for ( y = 0; y < srcH / 2 ; ++ y )
        {
            for ( x = 0; x < srcW / 2; ++ x ) 
            {
                if ( y & 1 ) {
                    b = (double)(*(pDestEighthU2 + iDstStrideUV * (y / 2 ) + x ));
                } else {
                    b = (double)(*(pDestEighthU1 + iDstStrideUV * (y / 2 ) + x ));
                }
                a = (double)(*(pDestU420 + y * iDstStrideUV + x));
                d = (double)(*(pDestHalfU + y * iDstStride + x * 2 + 1 ));
                c = (double)(*(pDestHalfU + y * iDstStride + x * 2 ));
                tmp = (( a + dAlpha * (b + c) + dBeta * d ) / ( 1 + dAlpha * 2 + dBeta ) + 0.5 );
                tmp = tmp > 255.0 ? 255 : tmp;
                tmp = tmp < 0.0 ? 0 : tmp;
                *(pDestU420 + y * iDstStrideUV + x) = (byte)tmp;
                if ( y & 1 ) {
                    b = (double)(*(pDestEighthV2 + iDstStrideUV * (y / 2 ) + x ));
                } else {
                    b = (double)(*(pDestEighthV1 + iDstStrideUV * (y / 2 ) + x ));
                }
                a = (double)(*(pDestV420 + y * iDstStrideUV + x));
                d = (double)(*(pDestHalfV + y * iDstStride + x * 2 + 1 ));
                c = (double)(*(pDestHalfV + y * iDstStride + x * 2 ));
                tmp = (( a + dAlpha * (b + c) + dBeta * d ) / ( 1 + dAlpha * 2 + dBeta ) + 0.5 );
                tmp = tmp > 255.0 ? 255 : tmp;
                tmp = tmp < 0.0 ? 0 : tmp;
                *(pDestV420 + y * iDstStrideUV + x) = (byte)tmp;
            }
        }
        // band separation
        pDestY420		= pDstY;
        pDestU420		= pDstU;
        pDestV420		= pDstV;
        pDestHalfU		= pDstY+iDstStride*srcH;
        pDestHalfV		= pDstY+iDstStride*srcH*3/2;
        pDestEighthU1	= pDstU+iDstStrideUV*srcH/2;
        pDestEighthV1	= pDstU+iDstStrideUV*srcH*3/4;
        pDestEighthU2	= pDstV+iDstStrideUV*srcH/2;
        pDestEighthV2	= pDstV+iDstStrideUV*srcH*3/4;
        for ( y = 0; y < srcH / 2; ++ y ) {
            for ( x = 0; x < srcW; ++ x ) {
                if ( !(x & 1) ) {
                    ppCU[y][x] = *(pDestU420 + iDstStrideUV * y + x / 2) + *(pDestHalfU + iDstStride * y + x);
                    ppCV[y][x] = *(pDestV420 + iDstStrideUV * y + x / 2) + *(pDestHalfV + iDstStride * y + x);
                    ppDU[y][x] = *(pDestU420 + iDstStrideUV * y + x / 2) - *(pDestHalfU + iDstStride * y + x);
                    ppDV[y][x] = *(pDestV420 + iDstStrideUV * y + x / 2) - *(pDestHalfV + iDstStride * y + x);
                } else {
                    if ( !(y & 1) ) {
                        ppCU[y][x] = *(pDestEighthU1 + iDstStrideUV * (y >> 1) + x / 2) + *(pDestHalfU + iDstStride * y + x);
                        ppCV[y][x] = *(pDestEighthV1 + iDstStrideUV * (y >> 1) + x / 2) + *(pDestHalfV + iDstStride * y + x);
                        ppDU[y][x] = *(pDestEighthU1 + iDstStrideUV * (y >> 1) + x / 2) - *(pDestHalfU + iDstStride * y + x);
                        ppDV[y][x] = *(pDestEighthV1 + iDstStrideUV * (y >> 1) + x / 2) - *(pDestHalfV + iDstStride * y + x);
                    } else {
                        ppCU[y][x] = *(pDestEighthU2 + iDstStrideUV * (y >> 1) + x / 2) + *(pDestHalfU + iDstStride * y + x);
                        ppCV[y][x] = *(pDestEighthV2 + iDstStrideUV * (y >> 1) + x / 2) + *(pDestHalfV + iDstStride * y + x);
                        ppDU[y][x] = *(pDestEighthU2 + iDstStrideUV * (y >> 1) + x / 2) - *(pDestHalfU + iDstStride * y + x);
                        ppDV[y][x] = *(pDestEighthV2 + iDstStrideUV * (y >> 1) + x / 2) - *(pDestHalfV + iDstStride * y + x);
                    }
                }
            }
        }
        for ( y = 0; y < srcH / 2; ++ y ) {
            for ( x = 0; x < srcW / 2; ++ x ) {
                ppCU[y][2 * x] = ppCU[y][2 * x] + ppCU[y][2 * x + 1];
                ppCV[y][2 * x] = ppCV[y][2 * x] + ppCV[y][2 * x + 1];
            }
            for ( x = 0; x < srcW / 2; ++ x ) {
                ppCU[y][2 * x + 1] = ppCU[y][2 * x] - 2 * ppCU[y][2 * x + 1];
                ppCV[y][2 * x + 1] = ppCV[y][2 * x] - 2 * ppCV[y][2 * x + 1];
            }
        }
        for ( y = 0; y < srcH / 2; ++ y ) {
            for ( x = 0; x < srcW / 2; ++ x ) {
                *(pDestU420 + y * iDstStrideUV + x) = ClipToByte( (ppCU[y][2 * x] + 2) / 4 );
                *(pDestV420 + y * iDstStrideUV + x) = ClipToByte( (ppCV[y][2 * x] + 2) / 4 );
            }
            for ( x = 0; x < srcW; ++ x ) {
                *(pDestHalfU + y * iDstStride + x) = ClipToByte( (ppDU[y][x] + 1) / 2 + 128 );
                *(pDestHalfV + y * iDstStride + x) = ClipToByte( (ppDV[y][x] + 1) / 2 + 128 );
            }
        }
        for ( y = 0; y < srcH / 4; ++ y ) {
            for ( x = 0; x < srcW / 2; ++ x ) {
                *(pDestEighthU1 + y * iDstStrideUV + x) = ClipToByte( ( ppCU[2 * y][2 * x + 1] + 2 ) / 4 + 128 );
                *(pDestEighthV1 + y * iDstStrideUV + x) = ClipToByte( ( ppCV[2 * y][2 * x + 1] + 2 ) / 4 + 128 );
                *(pDestEighthU2 + y * iDstStrideUV + x) = ClipToByte( ( ppCU[2 * y + 1][2 * x + 1] + 2 ) / 4 + 128 );
                *(pDestEighthV2 + y * iDstStrideUV + x) = ClipToByte( ( ppCV[2 * y + 1][2 * x + 1] + 2 ) / 4 + 128 );
            }
        }
        fwrite( pDstY, sizeof(byte), srcH * srcW * 2, pDstFile );
        fwrite( pDstU, sizeof(byte), srcH * srcW / 2, pDstFile );
        fwrite( pDstV, sizeof(byte), srcH * srcW / 2, pDstFile );
        iCnt ++;
        printf( "\b\b\b\b\b%04d.", iCnt );
    }
    printf( "\n" );
    for ( y = 0; y < srcH / 2; ++ y ) {
        delete [] ppCU[y];
        delete [] ppCV[y];
        delete [] ppDU[y];
        delete [] ppDV[y];
    }
    delete[] pSrc;
    delete[] pDstY;
    delete[] pDstU;
    delete[] pDstV;
    fclose( pSrcFile );
    fclose( pDstFile );
}

void FrameUnpacking( char *InputName, char *OutName, int srcH, int srcW, double dAlpha, double dBeta, FileType eFileType )
{
    FILE *pSrcFile = fopen( InputName, "rb" );
    FILE *pDstFile = fopen( OutName, "wb" );
    int x, y, Y, U, V, iFactorFileType;
    byte* pDstY, *pDstU, *pDstV, *pDstBGR;
    if ( RGB32 == eFileType )
    {
        iFactorFileType = 4;
        pDstBGR = new byte[srcH * srcW * 4];
    }
    else
    {
        iFactorFileType = 3;
        pDstY = new byte[srcH * srcW];
        pDstU = new byte[srcH * srcW];
        pDstV = new byte[srcH * srcW];
    }
    byte *pSrcYUV420 = new byte[srcH * srcW * iFactorFileType];
    int dwStrideSrcY = srcW;
    int dwStrideSrcUV = srcW / 2;
    int iCnt = 0;
    printf( "Finished:      " );
    while ( !feof( pSrcFile ) ) 
    {
        byte *pMY               = pDstY;
        byte *pMU               = pDstU;
        byte *pMV               = pDstV;
        byte *pSrc420Y          = pSrcYUV420;
        byte *pSrc420U          = pSrc420Y + srcH * srcW * 2;
        byte *pSrc420V          = pSrc420U + srcH * srcW / 2;
        byte *pSrcHalfU         = pSrc420Y + srcH * srcW; 
        byte *pSrcHalfV         = pSrc420Y + srcH * srcW * 3 / 2;
        byte *pSrcEighthU1	    = pSrc420U + srcH * dwStrideSrcUV / 2;
        byte *pSrcEighthV1	    = pSrc420U + srcH * dwStrideSrcUV * 3 / 4;
        byte *pSrcEighthU2	    = pSrc420V + srcH * dwStrideSrcUV / 2;
        byte *pSrcEighthV2	    = pSrc420V + srcH * dwStrideSrcUV * 3 / 4;
        fread( pSrcYUV420, sizeof(byte), srcH * srcW * 3, pSrcFile );
        if ( feof( pSrcFile ) )
        {
            break;
        }
        double a, b, c, d, tmp;
        for ( y = 0; y < srcH / 2 ; ++ y ) 
        {
            for ( x = 0; x < srcW / 2; ++ x ) 
            {
                if ( y & 1 ) {
                    b = (double)(*(pSrcEighthU2 + dwStrideSrcUV * (y / 2 ) + x ));
                } else {
                    b = (double)(*(pSrcEighthU1 + dwStrideSrcUV * (y / 2 ) + x ));
                }
                a = (double)(*(pSrc420U + y * dwStrideSrcUV + x));
                d = (double)(*(pSrcHalfU + y * dwStrideSrcY + x * 2 + 1 ));
                c = (double)(*(pSrcHalfU + y * dwStrideSrcY + x * 2 ));
                tmp = (( ( 1 + dAlpha * 2 + dBeta ) * a - dAlpha * (b + c) - dBeta * d ) + 0.5 );
                tmp = tmp > 255.0 ? 255 : tmp;
                tmp = tmp < 0.0 ? 0 : tmp;
                *(pSrc420U + y * dwStrideSrcUV + x) =  (byte)tmp;
                if ( y & 1 ) {
                    b = (double)(*(pSrcEighthV2 + dwStrideSrcUV * (y / 2 ) + x ));
                } else {
                    b = (double)(*(pSrcEighthV1 + dwStrideSrcUV * (y / 2 ) + x ));
                }
                a = (double)(*(pSrc420V + y * dwStrideSrcUV + x));
                d = (double)(*(pSrcHalfV + y * dwStrideSrcY + x * 2 + 1 ));
                c = (double)(*(pSrcHalfV + y * dwStrideSrcY + x * 2 ));
                tmp = (( ( 1 + dAlpha * 2 + dBeta ) * a - dAlpha * (b + c) - dBeta * d ) + 0.5 );
                tmp = tmp > 255.0 ? 255 : tmp;
                tmp = tmp < 0.0 ? 0 : tmp;
                *(pSrc420V + y * dwStrideSrcUV + x) = (byte)tmp;
            }
        }
        for ( y = 0; y < srcH; ++ y ) 
        {
            for ( x = 0; x < srcW; ++ x ) 
            {
                Y = *( pSrc420Y + x );
                if ( y & 1 ) {
                    U = *(pSrcHalfU + x); 
                    V = *(pSrcHalfV + x); 
                } else if ( 0==( x & 1 ) ) {
                    U = *(pSrc420U + x/2); 
                    V = *(pSrc420V + x/2 ); 
                } else if ( 1==(x & 1) && 0==(y & 2) ) {
                    U = *(pSrcEighthU1 + x/2); 
                    V = *(pSrcEighthV1 + x/2); 
                } else {//if ( 1==(dwW & 1) && (dwH & 2) )
                    U = *(pSrcEighthU2 + x/2); 
                    V = *(pSrcEighthV2 + x/2); 
                }
                if ( YUV444 == eFileType )
                {
                    *(pMY + x) = Y;
                    *(pMU + x) = U;
                    *(pMV + x) = V;
                }
                else
                {
                    InverseColorConversion( Y, U, V, pDstBGR + (y * srcW + x ) * 4 + 2, pDstBGR + (y * srcW + x ) * 4 + 1, pDstBGR + (y * srcW + x ) * 4 );
                }
            }
            if ( YUV444 == eFileType )
            {
                pMY += dwStrideSrcY;
                pMU += dwStrideSrcY;
                pMV += dwStrideSrcY;
            }
            pSrc420Y += dwStrideSrcY;
            if ( y & 1 ) {
                pSrcHalfU += dwStrideSrcY;
                pSrcHalfV += dwStrideSrcY;
            }
            else {
                pSrc420U    += dwStrideSrcUV;
                pSrc420V    += dwStrideSrcUV;
                if ( 0 == (y & 2) ) {
                    pSrcEighthU1 += dwStrideSrcUV;
                    pSrcEighthV1 += dwStrideSrcUV;
                }
                else {
                    pSrcEighthU2 += dwStrideSrcUV;
                    pSrcEighthV2 += dwStrideSrcUV;
                }
            }
        }
        if ( RGB32 == eFileType )
        {
            fwrite( pDstBGR, sizeof(byte), srcH * srcW * 4, pDstFile );
        }
        else
        {
            fwrite( pDstY, sizeof(byte), srcH * srcW, pDstFile );
            fwrite( pDstU, sizeof(byte), srcH * srcW, pDstFile );
            fwrite( pDstV, sizeof(byte), srcH * srcW, pDstFile );
        }
        iCnt ++;
        printf( "\b\b\b\b\b%04d.", iCnt );
    }
    printf( "\n" );
    delete [] pSrcYUV420;
    if ( RGB32 == eFileType )
    {
        delete [] pDstBGR;
    }
    else
    {
        delete [] pDstY;
        delete [] pDstU;
        delete [] pDstV;
    }
    fclose( pSrcFile );
    fclose( pDstFile );
}

void FrameUnpackingWithBandSeparation( char *InputName, char *OutName, int srcH, int srcW, double dAlpha, double dBeta, FileType eFileType ) 
{
    FILE *pSrcFile = fopen( InputName, "rb" );
    FILE *pDstFile = fopen( OutName, "wb" );
    int x, y, Y, U, V, tmp, iFactorFileType;
    byte* pDstY, *pDstU, *pDstV, *pDstBGR;
    if ( RGB32 == eFileType )
    {
        iFactorFileType = 4;
        pDstBGR = new byte[srcH * srcW * 4];
    }
    else
    {
        iFactorFileType = 3;
        pDstY = new byte[srcH * srcW];
        pDstU = new byte[srcH * srcW];
        pDstV = new byte[srcH * srcW];
    }
    byte *pSrcYUV420 = new byte[srcH * srcW * iFactorFileType];
    int dwStrideSrcY = srcW;
    int dwStrideSrcUV = srcW / 2;
    int **ppCU, **ppCV, **ppDU, **ppDV;
    ppCU = new int*[srcH / 2];
    ppDU = new int*[srcH / 2];
    ppCV = new int*[srcH / 2];
    ppDV = new int*[srcH / 2];
    for ( y = 0; y < srcH / 2; ++ y ) {
        ppCU[y] = new int[srcW];
        ppCV[y] = new int[srcW];
        ppDU[y] = new int[srcW];
        ppDV[y] = new int[srcW];
    }
    int iCnt = 0;
    printf( "Finished:      " );
    while ( !feof( pSrcFile ) )
    {
        byte *pMY               = pDstY;
        byte *pMU               = pDstU;
        byte *pMV               = pDstV;
        byte *pSrc420Y          = pSrcYUV420;
        byte *pSrc420U          = pSrc420Y + srcH * srcW * 2;
        byte *pSrc420V          = pSrc420U + srcH * srcW / 2;
        byte *pSrcHalfU         = pSrc420Y + srcH * srcW; 
        byte *pSrcHalfV         = pSrc420Y + srcH * srcW * 3 / 2;
        byte *pSrcEighthU1	    = pSrc420U + srcH * dwStrideSrcUV / 2;
        byte *pSrcEighthV1	    = pSrc420U + srcH * dwStrideSrcUV * 3 / 4;
        byte *pSrcEighthU2	    = pSrc420V + srcH * dwStrideSrcUV / 2;
        byte *pSrcEighthV2	    = pSrc420V + srcH * dwStrideSrcUV * 3 / 4;
        fread( pSrcYUV420, sizeof(byte), srcH * srcW * 3, pSrcFile );
        if ( feof( pSrcFile ) ) 
        {
            break;
        }
        // band de-separation
        for ( y = 0; y < srcH / 2; ++ y ) {
            for ( x = 0; x < srcW / 2; ++ x ) {
                tmp = *(pSrc420U + y * dwStrideSrcUV + x);
                ppCU[y][2 * x] = ( (int)tmp ) * 4;
                ppCV[y][2 * x] = (((int)(*(pSrc420V + y * dwStrideSrcUV + x )))  ) * 4;
            }
            for ( x = 0; x < srcW; ++ x ) {
                ppDU[y][x] = (((int)(*( pSrcHalfU + y * dwStrideSrcY + x))) - 128 ) * 2;
                ppDV[y][x] = (((int)(*( pSrcHalfV + y * dwStrideSrcY + x))) - 128 ) * 2;
            }
        }
        for ( y = 0; y < srcH / 4; ++ y ) {
            for ( x = 0; x < srcW / 2; ++ x ) {
                ppCU[2 * y][2 * x + 1] = ( ( (int)(*(pSrcEighthU1 + dwStrideSrcUV * y + x ) ) ) - 128 ) * 4;
                ppCV[2 * y][2 * x + 1] = ( ( (int)(*(pSrcEighthV1 + dwStrideSrcUV * y + x ) ) ) - 128 ) * 4;
                ppCU[2 * y + 1][2 * x + 1] = ( ( (int)(*(pSrcEighthU2 + dwStrideSrcUV * y + x ) ) ) - 128 ) * 4;
                ppCV[2 * y + 1][2 * x + 1] = ( ( (int)(*(pSrcEighthV2 + dwStrideSrcUV * y + x ) ) ) - 128 ) * 4;
            }
        }
        for ( y = 0; y < srcH / 2; ++ y ) {
            for ( x = 0; x < srcW / 2; ++ x ) {
                ppCU[y][2 * x] = ( ppCU[y][2 * x] + ppCU[y][2 * x + 1] ) / 2;
                ppCV[y][2 * x] = ( ppCV[y][2 * x] + ppCV[y][2 * x + 1] ) / 2;
            }
            for ( x = 0; x < srcW / 2; ++ x ) {
                ppCU[y][2 * x + 1] = ppCU[y][2 * x] - ppCU[y][2 * x + 1];
                ppCV[y][2 * x + 1] = ppCV[y][2 * x] - ppCV[y][2 * x + 1];
            }
        }
        for ( y = 0; y < srcH / 2; ++ y ) {
            for ( x = 0; x < srcW; ++ x ) {
                *(pSrcHalfU + dwStrideSrcY * y + x ) = ClipToByte( ( ppCU[y][x] - ppDU[y][x] ) / 2 );
                *(pSrcHalfV + dwStrideSrcY * y + x ) = ClipToByte( ( ppCV[y][x] - ppDV[y][x] ) / 2 );
                if ( !(x & 1) ) {
                    *(pSrc420U + dwStrideSrcUV * y + x / 2) = ClipToByte( (ppCU[y][x] + ppDU[y][x]) / 2 );
                    *(pSrc420V + dwStrideSrcUV * y + x / 2) = ClipToByte( (ppCV[y][x] + ppDV[y][x]) / 2 );
                } else {
                    if ( !(y & 1) ) {
                        *(pSrcEighthU1 + dwStrideSrcUV * (y >> 1) + x / 2 ) = ClipToByte( (ppCU[y][x] + ppDU[y][x]) / 2 );
                        *(pSrcEighthV1 + dwStrideSrcUV * (y >> 1) + x / 2 ) = ClipToByte( (ppCV[y][x] + ppDV[y][x]) / 2 );
                    } else {
                        *(pSrcEighthU2 + dwStrideSrcUV * (y >> 1) + x / 2 ) = ClipToByte( (ppCU[y][x] + ppDU[y][x]) / 2 );
                        *(pSrcEighthV2 + dwStrideSrcUV * (y >> 1) + x / 2 ) = ClipToByte( (ppCV[y][x] + ppDV[y][x]) / 2 );
                    }
                }
            }
        }
        // post filtering
        pSrc420Y          = pSrcYUV420;
        pSrc420U          = pSrc420Y + srcH * srcW * 2;
        pSrc420V          = pSrc420U + srcH * srcW / 2;
        pSrcHalfU         = pSrc420Y + srcH * srcW; 
        pSrcHalfV         = pSrc420Y + srcH * srcW * 3 / 2;
        pSrcEighthU1	    = pSrc420U + srcH * dwStrideSrcUV / 2;
        pSrcEighthV1	    = pSrc420U + srcH * dwStrideSrcUV * 3 / 4;
        pSrcEighthU2	    = pSrc420V + srcH * dwStrideSrcUV / 2;
        pSrcEighthV2	    = pSrc420V + srcH * dwStrideSrcUV * 3 / 4;
        double a, b, c, d, tmp;
        for ( y = 0; y < srcH / 2 ; ++ y ) {
            for ( x = 0; x < srcW / 2; ++ x ) {
                if ( y & 1 ) {
                    b = (double)(*(pSrcEighthU2 + dwStrideSrcUV * (y / 2 ) + x ));
                } else {
                    b = (double)(*(pSrcEighthU1 + dwStrideSrcUV * (y / 2 ) + x ));
                }
                a = (double)(*(pSrc420U + y * dwStrideSrcUV + x));
                d = (double)(*(pSrcHalfU + y * dwStrideSrcY + x * 2 + 1 ));
                c = (double)(*(pSrcHalfU + y * dwStrideSrcY + x * 2 ));
                tmp = (( ( 1 + dAlpha * 2 + dBeta ) * a - dAlpha * (b + c) - dBeta * d ) + 0.5 );
                tmp = tmp > 255.0 ? 255 : tmp;
                tmp = tmp < 0.0 ? 0 : tmp;
                *(pSrc420U + y * dwStrideSrcUV + x) =  (byte)tmp;
                if ( y & 1 ) {
                    b = (double)(*(pSrcEighthV2 + dwStrideSrcUV * (y / 2 ) + x ));
                } else {
                    b = (double)(*(pSrcEighthV1 + dwStrideSrcUV * (y / 2 ) + x ));
                }
                a = (double)(*(pSrc420V + y * dwStrideSrcUV + x));
                d = (double)(*(pSrcHalfV + y * dwStrideSrcY + x * 2 + 1 ));
                c = (double)(*(pSrcHalfV + y * dwStrideSrcY + x * 2 ));
                tmp = (( ( 1 + dAlpha * 2 + dBeta ) * a - dAlpha * (b + c) - dBeta * d ) + 0.5 );
                tmp = tmp > 255.0 ? 255 : tmp;
                tmp = tmp < 0.0 ? 0 : tmp;
                *(pSrc420V + y * dwStrideSrcUV + x) = (byte)tmp;
            }
        }
        for ( y = 0; y < srcH; ++ y ) {
            for ( x = 0; x < srcW; ++ x ) {
                Y = *( pSrc420Y + x );
                if ( y & 1 ) {
                    U = *(pSrcHalfU + x); 
                    V = *(pSrcHalfV + x); 
                } else if ( 0==( x & 1 ) ) {
                    U = *(pSrc420U + x/2); 
                    V = *(pSrc420V + x/2 ); 
                } else if ( 1==(x & 1) && 0==(y & 2) ) {
                    U = *(pSrcEighthU1 + x/2); 
                    V = *(pSrcEighthV1 + x/2); 
                } else {//if ( 1==(dwW & 1) && (dwH & 2) )
                    U = *(pSrcEighthU2 + x/2); 
                    V = *(pSrcEighthV2 + x/2); 
                }
                if ( YUV444 == eFileType )
                {
                    *(pMY + x) = Y;
                    *(pMU + x) = U;
                    *(pMV + x) = V;
                }
                else
                {
                    InverseColorConversion( Y, U, V, pDstBGR + (y * srcW + x ) * 4 + 2, pDstBGR + (y * srcW + x ) * 4 + 1, pDstBGR + (y * srcW + x ) * 4 );
                }
            }
            if ( YUV444 == eFileType )
            {
                pMY += dwStrideSrcY;
                pMU += dwStrideSrcY;
                pMV += dwStrideSrcY;
            }
            pSrc420Y += dwStrideSrcY;
            if ( y & 1 ) {
                pSrcHalfU += dwStrideSrcY;
                pSrcHalfV += dwStrideSrcY;
            }
            else {
                pSrc420U    += dwStrideSrcUV;
                pSrc420V    += dwStrideSrcUV;
                if ( 0 == (y & 2) ) {
                    pSrcEighthU1 += dwStrideSrcUV;
                    pSrcEighthV1 += dwStrideSrcUV;
                }
                else {
                    pSrcEighthU2 += dwStrideSrcUV;
                    pSrcEighthV2 += dwStrideSrcUV;
                }
            }
        }
        if ( RGB32 == eFileType )
        {
            fwrite( pDstBGR, sizeof(byte), srcH * srcW * 4, pDstFile );
        }
        else
        {
            fwrite( pDstY, sizeof(byte), srcH * srcW, pDstFile );
            fwrite( pDstU, sizeof(byte), srcH * srcW, pDstFile );
            fwrite( pDstV, sizeof(byte), srcH * srcW, pDstFile );
        }
        iCnt ++;
        printf( "\b\b\b\b\b%04d.", iCnt );
    }
    printf( "\n" );
    for ( y = 0; y < srcH / 2; ++ y ) {
        delete [] ppCU[y];
        delete [] ppCV[y];
        delete [] ppDU[y];
        delete [] ppDV[y];
    }
    delete ppCU;
    delete ppDU;
    delete ppCV;
    delete ppDV;
    delete [] pSrcYUV420;
    if ( RGB32 == eFileType )
    {
        delete [] pDstBGR;
    }
    else
    {
        delete [] pDstY;
        delete [] pDstU;
        delete [] pDstV;
    }
    fclose( pSrcFile );
    fclose( pDstFile );
}

void RGB32ToYUV444( char *InputName, char *OutName, int srcH, int srcW ) {
    FILE *pSrcFile = fopen( InputName, "rb" );
    FILE *pDstFile = fopen( OutName, "wb" );
    byte *pSrcRGB32 = new byte[srcH * srcW * 4];
    byte *pDstY = new byte[srcH * srcW];
    byte *pDstU = new byte[srcH * srcW];
    byte *pDstV = new byte[srcH * srcW];
    int iDstStride = srcW;


    int x, y, R, G, B;
    byte Y, U, V;
    int iCnt = 0;
    printf( "Finished:      " );
    while ( !feof( pSrcFile ) ) 
    {
        byte *pSrcRGB       = pSrcRGB32;
        byte *pDestY420		= pDstY;
        byte *pDestU420		= pDstU;
        byte *pDestV420		= pDstV;
        fread( pSrcRGB32, sizeof(byte), srcH * srcW * 4, pSrcFile );
        if ( feof( pSrcFile ) )
        {
            break;
        }
        for ( y = 0; y < srcH; ++ y ) 
        {
            for ( x = 0; x < srcW; ++ x ) 
            {
                B =  pSrcRGB[4*x + 0];
                G =  pSrcRGB[4*x + 1];
                R =  pSrcRGB[4*x + 2];
                ColorConversion( R, G, B, &Y, &U, &V );
                *(pDestY420 + x) = Y; 
                *(pDestU420 + x) = U; 
                *(pDestV420 + x) = V; 
            }
            pSrcRGB += iDstStride * 4;
            pDestY420 += iDstStride;
            pDestU420 += iDstStride;
            pDestV420 += iDstStride;
        }
        fwrite( pDstY, sizeof(byte), srcH * srcW, pDstFile );
        fwrite( pDstU, sizeof(byte), srcH * srcW, pDstFile );
        fwrite( pDstV, sizeof(byte), srcH * srcW, pDstFile );
        iCnt ++;
        printf( "\b\b\b\b\b%04d.", iCnt );
    }
    printf( "\n" );
    delete[] pSrcRGB32;
    delete[] pDstY;
    delete[] pDstU;
    delete[] pDstV;
    fclose( pSrcFile );
    fclose( pDstFile );
}

void YUVHorizontalSplit( char *InputName, char *OutName, int srcH, int srcW, int iNumFrames, YUVFormat InFormat ) {
    char rgNameBuff[FILENAME_MAX];
    strcpy( rgNameBuff, InputName );
    strcat( rgNameBuff, ".yuv" );
    FILE *pSrcFile = fopen( rgNameBuff, "rb" );
    strcpy( rgNameBuff, OutName );
    strcat( rgNameBuff, "_top.yuv" );
    FILE *pDstFileTop = fopen( rgNameBuff, "wb" );
    strcpy( rgNameBuff, OutName );
    strcat( rgNameBuff, "_bot.yuv" );
    FILE *pDstFileBottom = fopen( rgNameBuff, "wb" );
    byte *pY = new byte[srcW * srcH];
    byte *pU = new byte[srcW * srcH >> 2];
    byte *pV = new byte[srcW * srcH >> 2];

    for ( int i = 0; i < iNumFrames; ++ i ) {
        fread( pY, sizeof(byte), srcH * srcW, pSrcFile );
        fread( pU, sizeof(byte), srcH * srcW >> 2, pSrcFile );
        fread( pV, sizeof(byte), srcH * srcW >> 2, pSrcFile );
        fwrite( pY, sizeof(byte), srcH * srcW >> 1, pDstFileTop );
        fwrite( pY + srcH * srcW / 2, sizeof(byte), srcH * srcW >> 1, pDstFileBottom );
        fwrite( pU, sizeof(byte), srcH * srcW >> 3, pDstFileTop );
        fwrite( pV, sizeof(byte), srcH * srcW >> 3, pDstFileTop );
        fwrite( pU + srcH * srcW / 8, sizeof(byte), srcH * srcW >> 3, pDstFileBottom );
        fwrite( pV + srcH * srcW / 8, sizeof(byte), srcH * srcW >> 3, pDstFileBottom );
    }
    delete [] pY;
    delete [] pU;
    delete [] pV;
    fclose( pSrcFile );
    fclose( pDstFileTop );
    fclose( pDstFileBottom );
}

void YUVHorizontalFuse( char *InputNameTop, char *InputNameBot, char *OutName, int srcH, int srcW, int iNumFrames, YUVFormat InFormat ) {
    char rgNameBuff[FILENAME_MAX];
    strcpy( rgNameBuff, InputNameTop );
    strcat( rgNameBuff, ".yuv" );
    FILE *pSrcFileTop = fopen( rgNameBuff, "rb" );
    strcpy( rgNameBuff, InputNameBot );
    strcat( rgNameBuff, ".yuv" );
    FILE *pSrcFileBottom = fopen( rgNameBuff, "rb" );
    strcpy( rgNameBuff, OutName );
    strcat( rgNameBuff, ".yuv" );
    FILE *pDstFile = fopen( rgNameBuff, "wb" );
    byte *pY = new byte[srcW * srcH];
    byte *pU = new byte[srcW * srcH >> 2];
    byte *pV = new byte[srcW * srcH >> 2];

    for ( int i = 0; i < iNumFrames; ++ i ) {
        fread( pY, sizeof(byte), srcH * srcW >> 1, pSrcFileTop );
        fread( pY + srcH * srcW / 2, sizeof(byte), srcH * srcW >> 1, pSrcFileBottom );
        fread( pU, sizeof(byte), srcH * srcW >> 3, pSrcFileTop );
        fread( pV, sizeof(byte), srcH * srcW >> 3, pSrcFileTop );
        fread( pU + srcH * srcW / 8, sizeof(byte), srcH * srcW >> 3, pSrcFileBottom );
        fread( pV + srcH * srcW / 8, sizeof(byte), srcH * srcW >> 3, pSrcFileBottom );
        fwrite( pY, sizeof(byte), srcH * srcW, pDstFile );
        fwrite( pU, sizeof(byte), srcH * srcW >> 2, pDstFile );
        fwrite( pV, sizeof(byte), srcH * srcW >> 2, pDstFile );
    }
    delete [] pY;
    delete [] pU;
    delete [] pV;
    fclose( pDstFile );
    fclose( pSrcFileTop );
    fclose( pSrcFileBottom );
}

void FramePackingWithLiftedBandSeparation( char *InputName, char *OutName, int srcH, int srcW, YUVFormat InFormat, double dAlpha, double dBeta ) {
    FILE *pSrcFile = fopen( InputName, "rb" );
    FILE *pDstFile = fopen( OutName, "wb" );
    byte *pSrcRGB32 = new byte[srcH * srcW * 4];
    byte *pDstY = new byte[srcH * srcW * 2];
    byte *pDstU = new byte[srcH * srcW / 2];
    byte *pDstV = new byte[srcH * srcW / 2];
    int iDstStride = srcW;
    int iDstStrideUV = srcW / 2;
    int i, x, y, R, G, B, Ytmp, Utmp, Vtmp;
    byte Y, U, V;
    int **ppCU, **ppCV, **ppDU, **ppDV;
    ppCU = new int*[srcH/ 2];
    ppDU = new int*[srcH / 2];
    ppCV = new int*[srcH / 2];
    ppDV = new int*[srcH / 2];
    for ( y = 0; y < srcH / 2; ++ y ) {
        ppCU[y] = new int[srcW];
        ppCV[y] = new int[srcW];
        ppDU[y] = new int[srcW];
        ppDV[y] = new int[srcW];
    }

    for ( i = 0; i < 10; ++ i ) {
        byte *pSrcRGB       = pSrcRGB32;
        byte *pDestY420		= pDstY;
        byte *pDestU420		= pDstU;
        byte *pDestV420		= pDstV;
        byte *pDestHalfU    = pDstY+iDstStride*srcH;
        byte *pDestHalfV	= pDstY+iDstStride*srcH*3/2;
        byte *pDestEighthU1	= pDstU+iDstStrideUV*srcH/2;
        byte *pDestEighthV1	= pDstU+iDstStrideUV*srcH*3/4;
        byte *pDestEighthU2	= pDstV+iDstStrideUV*srcH/2;
        byte *pDestEighthV2	= pDstV+iDstStrideUV*srcH*3/4;
        fread( pSrcRGB32, sizeof(byte), srcH * srcW * 4, pSrcFile );
        for ( y = 0; y < srcH; ++ y ) {
            for ( x = 0; x < srcW; ++ x ) {
                B =  pSrcRGB[4*x + 0];
                G =  pSrcRGB[4*x + 1];
                R =  pSrcRGB[4*x + 2];
                Ytmp = ( (  66 * R + 129 * G +  25 * B + 128) >> 8) +  16;
                Utmp = ( ( -38 * R -  74 * G + 112 * B + 128) >> 8) + 128;
                Vtmp = ( ( 112 * R -  94 * G -  18 * B + 128) >> 8) + 128;
                Y = (byte)((Ytmp < 0) ? 0 : ((Ytmp > 255) ? 255 : Ytmp));
                U = (byte)((Utmp < 0) ? 0 : ((Utmp > 255) ? 255 : Utmp));
                V = (byte)((Vtmp < 0) ? 0 : ((Vtmp > 255) ? 255 : Vtmp));
                *(pDestY420 + x) = Y; 
                if ( y & 1 ) {
                    *(pDestHalfU+x) = U;
                    *(pDestHalfV+x) = V;
                } else if ( 0==(x & 1) ) {
                    *(pDestU420+x/2) = U;
                    *(pDestV420+x/2) = V;
                } else if ( 1== (x & 1 ) && 0==(y & 2) ) {
                    *(pDestEighthU1+x/2) = U;
                    *(pDestEighthV1+x/2) = V;
                } else { // if ( 1== (x & 1 ) && (y & 2) ) 
                    *(pDestEighthU2+x/2) = U;
                    *(pDestEighthV2+x/2) = V;
                }
            }
            pSrcRGB += iDstStride * 4;
            pDestY420 += iDstStride;
            if ( y & 1 ) {
                pDestHalfU  += iDstStride;
                pDestHalfV  += iDstStride;
            } else  {
                pDestU420 += iDstStrideUV;
                pDestV420 += iDstStrideUV;
                if ( 0== ( y & 2 ) ) {
                    pDestEighthU1 += iDstStrideUV;
                    pDestEighthV1 += iDstStrideUV;
                }
                else {
                    pDestEighthU2 += iDstStrideUV;
                    pDestEighthV2 += iDstStrideUV;
                }
            }
        }
        // band separation
        pDestY420		= pDstY;
        pDestU420		= pDstU;
        pDestV420		= pDstV;
        pDestHalfU		= pDstY+iDstStride*srcH;
        pDestHalfV		= pDstY+iDstStride*srcH*3/2;
        pDestEighthU1	= pDstU+iDstStrideUV*srcH/2;
        pDestEighthV1	= pDstU+iDstStrideUV*srcH*3/4;
        pDestEighthU2	= pDstV+iDstStrideUV*srcH/2;
        pDestEighthV2	= pDstV+iDstStrideUV*srcH*3/4;
        for ( y = 0; y < srcH / 2; ++ y ) {
            for ( x = 0; x < srcW; ++ x ) {
                if ( !(x & 1) ) {
                    ppCU[y][x] = *(pDestU420 + iDstStrideUV * y + x / 2) + *(pDestHalfU + iDstStride * y + x);
                    ppCV[y][x] = *(pDestV420 + iDstStrideUV * y + x / 2) + *(pDestHalfV + iDstStride * y + x);
                    ppDU[y][x] = *(pDestU420 + iDstStrideUV * y + x / 2) - *(pDestHalfU + iDstStride * y + x);
                    ppDV[y][x] = *(pDestV420 + iDstStrideUV * y + x / 2) - *(pDestHalfV + iDstStride * y + x);
                } else {
                    if ( !(y & 1) ) {
                        ppCU[y][x] = *(pDestEighthU1 + iDstStrideUV * (y >> 1) + x / 2) + *(pDestHalfU + iDstStride * y + x);
                        ppCV[y][x] = *(pDestEighthV1 + iDstStrideUV * (y >> 1) + x / 2) + *(pDestHalfV + iDstStride * y + x);
                        ppDU[y][x] = *(pDestEighthU1 + iDstStrideUV * (y >> 1) + x / 2) - *(pDestHalfU + iDstStride * y + x);
                        ppDV[y][x] = *(pDestEighthV1 + iDstStrideUV * (y >> 1) + x / 2) - *(pDestHalfV + iDstStride * y + x);
                    } else {
                        ppCU[y][x] = *(pDestEighthU2 + iDstStrideUV * (y >> 1) + x / 2) + *(pDestHalfU + iDstStride * y + x);
                        ppCV[y][x] = *(pDestEighthV2 + iDstStrideUV * (y >> 1) + x / 2) + *(pDestHalfV + iDstStride * y + x);
                        ppDU[y][x] = *(pDestEighthU2 + iDstStrideUV * (y >> 1) + x / 2) - *(pDestHalfU + iDstStride * y + x);
                        ppDV[y][x] = *(pDestEighthV2 + iDstStrideUV * (y >> 1) + x / 2) - *(pDestHalfV + iDstStride * y + x);
                    }
                }
            }
        }
        for ( y = 0; y < srcH / 2; ++ y ) {
            for ( x = 0; x < srcW / 2; ++ x ) {
                ppCU[y][2 * x] = ppCU[y][2 * x] + ppCU[y][2 * x + 1];
                ppCV[y][2 * x] = ppCV[y][2 * x] + ppCV[y][2 * x + 1];
            }
            for ( x = 0; x < srcW / 2; ++ x ) {
                ppCU[y][2 * x + 1] = ppCU[y][2 * x] - 2 * ppCU[y][2 * x + 1];
                ppCV[y][2 * x + 1] = ppCV[y][2 * x] - 2 * ppCV[y][2 * x + 1];
            }
        }
        for ( y = 0; y < srcH / 2; ++ y ) {
            for ( x = 0; x < srcW / 2; ++ x ) {
                *(pDestU420 + y * iDstStrideUV + x) = ClipToByte( ppCU[y][2 * x] >> 2 );
                *(pDestV420 + y * iDstStrideUV + x) = ClipToByte( ppCV[y][2 * x] >> 2 );
            }
            for ( x = 0; x < srcW; ++ x ) {
                *(pDestHalfU + y * iDstStride + x) = ClipToByte( ppDU[y][x] + 128 );
                *(pDestHalfV + y * iDstStride + x) = ClipToByte( ppDV[y][x] + 128 );
            }
        }
        for ( y = 0; y < srcH / 4; ++ y ) {
            for ( x = 0; x < srcW / 2; ++ x ) {
                *(pDestEighthU1 + y * iDstStrideUV + x) = ClipToByte( ppCU[2 * y][2 * x + 1] + 128 );
                *(pDestEighthV1 + y * iDstStrideUV + x) = ClipToByte( ppCV[2 * y][2 * x + 1] + 128 );
                *(pDestEighthU2 + y * iDstStrideUV + x) = ClipToByte( ppCU[2 * y + 1][2 * x + 1] + 128 );
                *(pDestEighthV2 + y * iDstStrideUV + x) = ClipToByte( ppCV[2 * y + 1][2 * x + 1] + 128 );
            }
        }
        fwrite( pDstY, sizeof(byte), srcH * srcW * 2, pDstFile );
        fwrite( pDstU, sizeof(byte), srcH * srcW / 2, pDstFile );
        fwrite( pDstV, sizeof(byte), srcH * srcW / 2, pDstFile );
        printf( "Frame %03d.\n", i );
    }
    for ( y = 0; y < srcH / 2; ++ y ) {
        delete [] ppCU[y];
        delete [] ppCV[y];
        delete [] ppDU[y];
        delete [] ppDV[y];
    }
    delete[] pSrcRGB32;
    delete[] pDstY;
    delete[] pDstU;
    delete[] pDstV;
    fclose( pSrcFile );
    fclose( pDstFile );
}

void FrameUnpackingWithLiftedBandSeparation( char *InputName, char *OutName, int srcH, int srcW, YUVFormat InFormat, double dAlpha, double dBeta ) {
       FILE *pSrcFile = fopen( InputName, "rb" );
    FILE *pDstFile = fopen( OutName, "wb" );
    int i, x, y, U, V, tmp;
    byte *pSrcYUV420 = new byte[srcH * srcW * 3];
    byte *pDstY = new byte[srcH * srcW];
    byte *pDstU = new byte[srcH * srcW];
    byte *pDstV = new byte[srcH * srcW];
    int dwStrideSrcY = srcW;
    int dwStrideSrcUV = srcW / 2;
    int **ppCU, **ppCV, **ppDU, **ppDV;
    ppCU = new int*[srcH / 2];
    ppDU = new int*[srcH / 2];
    ppCV = new int*[srcH / 2];
    ppDV = new int*[srcH / 2];
    for ( y = 0; y < srcH / 2; ++ y ) {
        ppCU[y] = new int[srcW];
        ppCV[y] = new int[srcW];
        ppDU[y] = new int[srcW];
        ppDV[y] = new int[srcW];
    }
    for ( i = 0; i < 10; ++ i ) {
        byte *pMY               = pDstY;
        byte *pMU               = pDstU;
        byte *pMV               = pDstV;
        byte *pSrc420Y          = pSrcYUV420;
        byte *pSrc420U          = pSrc420Y + srcH * srcW * 2;
        byte *pSrc420V          = pSrc420U + srcH * srcW / 2;
        byte *pSrcHalfU         = pSrc420Y + srcH * srcW; 
        byte *pSrcHalfV         = pSrc420Y + srcH * srcW * 3 / 2;
        byte *pSrcEighthU1	    = pSrc420U + srcH * dwStrideSrcUV / 2;
        byte *pSrcEighthV1	    = pSrc420U + srcH * dwStrideSrcUV * 3 / 4;
        byte *pSrcEighthU2	    = pSrc420V + srcH * dwStrideSrcUV / 2;
        byte *pSrcEighthV2	    = pSrc420V + srcH * dwStrideSrcUV * 3 / 4;
        fread( pSrcYUV420, sizeof(byte), srcH * srcW * 3, pSrcFile );
        // band separation
        for ( y = 0; y < srcH / 2; ++ y ) {
            for ( x = 0; x < srcW / 2; ++ x ) {
                tmp = *(pSrc420U + y * dwStrideSrcUV + x);
                ppCU[y][2 * x] = ( (int)tmp ) * 4;
                ppCV[y][2 * x] = (((int)(*(pSrc420V + y * dwStrideSrcUV + x )))  ) * 4;
            }
            for ( x = 0; x < srcW; ++ x ) {
                ppDU[y][x] = ((int)(*( pSrcHalfU + y * dwStrideSrcY + x))) - 128;
                ppDV[y][x] = ((int)(*( pSrcHalfV + y * dwStrideSrcY + x))) - 128;
            }
        }
        for ( y = 0; y < srcH / 4; ++ y ) {
            for ( x = 0; x < srcW / 2; ++ x ) {
                ppCU[2 * y][2 * x + 1] = ( (int)(*(pSrcEighthU1 + dwStrideSrcUV * y + x ) ) ) - 128;
                ppCV[2 * y][2 * x + 1] = ( (int)(*(pSrcEighthV1 + dwStrideSrcUV * y + x ) ) ) - 128;
                ppCU[2 * y + 1][2 * x + 1] = ( (int)(*(pSrcEighthU2 + dwStrideSrcUV * y + x ) ) ) - 128;
                ppCV[2 * y + 1][2 * x + 1] = ( (int)(*(pSrcEighthV2 + dwStrideSrcUV * y + x ) ) ) - 128;
            }
        }
        for ( y = 0; y < srcH / 2; ++ y ) {
            for ( x = 0; x < srcW / 2; ++ x ) {
                ppCU[y][2 * x] = ( ppCU[y][2 * x] + ppCU[y][2 * x + 1] ) / 2;
                ppCV[y][2 * x] = ( ppCV[y][2 * x] + ppCV[y][2 * x + 1] ) / 2;
            }
            for ( x = 0; x < srcW / 2; ++ x ) {
                ppCU[y][2 * x + 1] = ppCU[y][2 * x] - ppCU[y][2 * x + 1];
                ppCV[y][2 * x + 1] = ppCV[y][2 * x] - ppCV[y][2 * x + 1];
            }
        }
        for ( y = 0; y < srcH / 2; ++ y ) {
            for ( x = 0; x < srcW; ++ x ) {
                *(pSrcHalfU + dwStrideSrcY * y + x ) = ClipToByte( ( ppCU[y][x] - ppDU[y][x] ) / 2 );
                *(pSrcHalfV + dwStrideSrcY * y + x ) = ClipToByte( ( ppCV[y][x] - ppDV[y][x] ) / 2 );
                if ( !(x & 1) ) {
                    *(pSrc420U + dwStrideSrcUV * y + x / 2) = ClipToByte( (ppCU[y][x] + ppDU[y][x]) / 2 );
                    *(pSrc420V + dwStrideSrcUV * y + x / 2) = ClipToByte( (ppCV[y][x] + ppDV[y][x]) / 2 );
                } else {
                    if ( !(y & 1) ) {
                        *(pSrcEighthU1 + dwStrideSrcUV * (y >> 1) + x / 2 ) = ClipToByte( (ppCU[y][x] + ppDU[y][x]) / 2 );
                        *(pSrcEighthV1 + dwStrideSrcUV * (y >> 1) + x / 2 ) = ClipToByte( (ppCV[y][x] + ppDV[y][x]) / 2 );
                    } else {
                        *(pSrcEighthU2 + dwStrideSrcUV * (y >> 1) + x / 2 ) = ClipToByte( (ppCU[y][x] + ppDU[y][x]) / 2 );
                        *(pSrcEighthV2 + dwStrideSrcUV * (y >> 1) + x / 2 ) = ClipToByte( (ppCV[y][x] + ppDV[y][x]) / 2 );
                    }
                }
            }
        }

        for ( y = 0; y < srcH; ++ y ) {
            for ( x = 0; x < srcW; ++ x ) {
                *(pMY + x) = *( pSrc420Y + x );
                if ( y & 1 ) {
                    U = *(pSrcHalfU + x); 
                    V = *(pSrcHalfV + x); 
                } else if ( 0==( x & 1 ) ) {
                    U = *(pSrc420U + x/2); 
                    V = *(pSrc420V + x/2 ); 
                } else if ( 1==(x & 1) && 0==(y & 2) ) {
                    U = *(pSrcEighthU1 + x/2); 
                    V = *(pSrcEighthV1 + x/2); 
                } else {//if ( 1==(dwW & 1) && (dwH & 2) )
                    U = *(pSrcEighthU2 + x/2); 
                    V = *(pSrcEighthV2 + x/2); 
                }
                *(pMU + x) = U;
                *(pMV + x) = V;
            }
            pMY += dwStrideSrcY;
            pMU += dwStrideSrcY;
            pMV += dwStrideSrcY;
            pSrc420Y += dwStrideSrcY;
            if ( y & 1 ) {
                pSrcHalfU += dwStrideSrcY;
                pSrcHalfV += dwStrideSrcY;
            }
            else {
                pSrc420U    += dwStrideSrcUV;
                pSrc420V    += dwStrideSrcUV;
                if ( 0 == (y & 2) ) {
                    pSrcEighthU1 += dwStrideSrcUV;
                    pSrcEighthV1 += dwStrideSrcUV;
                }
                else {
                    pSrcEighthU2 += dwStrideSrcUV;
                    pSrcEighthV2 += dwStrideSrcUV;
                }
            }
        }
        fwrite( pDstY, sizeof(byte), srcH * srcW, pDstFile );
        fwrite( pDstU, sizeof(byte), srcH * srcW, pDstFile );
        fwrite( pDstV, sizeof(byte), srcH * srcW, pDstFile );
        printf( "Frame %03d.\n", i );
    }
    for ( y = 0; y < srcH / 2; ++ y ) {
        delete [] ppCU[y];
        delete [] ppCV[y];
        delete [] ppDU[y];
        delete [] ppDV[y];
    }
    delete ppCU;
    delete ppDU;
    delete ppCV;
    delete ppDV;
    delete [] pSrcYUV420;
    delete [] pDstY;
    delete [] pDstU;
    delete [] pDstV;
    fclose( pSrcFile );
    fclose( pDstFile );
}